/** ADAMEm: Coleco ADAM emulator ********************************************/
/**                                                                        **/
/**                               eosdisk.c                                **/
/**                                                                        **/
/** Various ADAM disk/tape image utilities                                 **/
/**                                                                        **/
/** Copyright (C) Marcel de Kogel 1996,1997                                **/
/**     You are not allowed to distribute this software commercially       **/
/**     Please, notify me, if you make any changes to this file            **/
/****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

typedef unsigned char byte;

#define imagesizemask   0xFFFF          /* Mask off upper two bytes of      */
                                        /* block numbers. Workaround to a   */
                                        /* PowerPaint bug                   */
#define MAX_FILE        (1024/26)       /* Maximum nr. of files per         */
                                        /* directory block                  */
#define MAX_DIR         (MAX_FILE*128)  /* Maximum nr. of files per         */
                                        /* directory                        */

static int tapeinterleave[]=            /* Interleave table for tape images */
{ 0,1,2,3,4,5,6,7 };
static int diskinterleave[]=            /* Interleave table for disk images */
{ 0,5,2,7,4,1,6,3 };
static int *interleave;                 /* Interleave table pointer         */
static FILE *infile,*outfile;           /* Input and output streams         */
static char tempfile[L_tmpnam];         /* Temporary file name              */
static int ignore_bytes_in_last_block=0;
static int include_system=0,include_deleted=0,keep_reserved=0;
static byte buffer[1024];

static byte newdisk[2048]=              /* First two blocks of an empty     */
{                                       /* 160K EOS disk/tape image         */
 #include "newdisk.h"
};

typedef struct
{
 char filename[13];                     /* file name only                   */
 char realname[12];                     /* file name + type + trailing $03  */
 char type;
 int attr;
 int startblock;
 int reservedblocks;
 int usedblocks;
 int bytesinlastblock;
 int day,month,year;
 int direntry;
} fileinfo_t;

static byte volume_info[26];            /* First 26 bytes of the directory  */
static fileinfo_t fileinfo[MAX_DIR];
static int dirsize,volumesize,first_free_dir_entry,first_free_block;
static int day,month,year;              /* System date                      */

/****************************************************************************/
/* Close all open streams and quit program                                  */
/****************************************************************************/
static void fatal (int code)
{
 if (infile) fclose (infile);
 if (outfile && outfile!=infile) fclose (outfile);
 if (tempfile[0]) remove (tempfile);
 exit (code);
}

/****************************************************************************/
/* Read given block nr. into given buffer                                   */
/****************************************************************************/
static int ReadBlock (int nr,byte *buf,FILE *f)
{
 nr*=2;
 if (fseek(f,((nr&(~7))+interleave[nr&7])*512,SEEK_SET))
  return -1;
 if (!fread(buf,512,1,f))
  return -2;
 nr+=1;
 if (fseek(f,((nr&(~7))+interleave[nr&7])*512,SEEK_SET))
  return -1;
 if (!fread(buf+512,512,1,f))
  return -2;
 return 0;
}

/****************************************************************************/
/* Write given block nr. from given buffer                                  */
/****************************************************************************/
static int WriteBlock (int nr,byte *buf,FILE *f)
{
 nr*=2;
 if (fseek(f,((nr&(~7))+interleave[nr&7])*512,SEEK_SET))
  return -1;
 if (!fwrite(buf,512,1,f))
  return -2;
 nr+=1;
 if (fseek(f,((nr&(~7))+interleave[nr&7])*512,SEEK_SET))
  return -1;
 if (!fwrite(buf+512,512,1,f))
  return -2;
 return 0;
}

/****************************************************************************/
/* Check if given filetype is a valid one                                   */
/****************************************************************************/
static int CheckFileType (char c)
{
 int filetypes[]= { 'c','C','a','A','h','H',0,1,2,-1 };
 int i;
 for (i=0;filetypes[i]>=0;++i) if (c==filetypes[i]) return 1;
 return 0;
}

/****************************************************************************/
/* Parse directory entry and fill fileinfo structure                        */
/****************************************************************************/
static void GetFileInfo (int nr,byte *p)
{
 int i;
 memset (&fileinfo[nr],0,sizeof(fileinfo[nr]));
 memcpy (fileinfo[nr].realname,p,12);
 for (i=0;i<12&&p[i]!=0x03;++i)
  fileinfo[nr].filename[i]=p[i];
 if (i)
 {
  if (CheckFileType(p[i-1]))
  {
   fileinfo[nr].filename[i-1]='\0';
   fileinfo[nr].type=p[i-1];
  }
  else fileinfo[nr].type=-1;
 }
 fileinfo[nr].attr=p[12];
 i=p[13]+p[14]*256+p[15]*256*256+p[16]*256*256*256;
 i&=imagesizemask;
 fileinfo[nr].startblock=i;
 i=p[17]+p[18]*256;
 i&=imagesizemask;
 fileinfo[nr].reservedblocks=i;
 i=p[19]+p[20]*256;
 i&=imagesizemask;
 fileinfo[nr].usedblocks=i;
 i=p[21]+p[22]*256;
 if (!i || i>1024) i=1024;
 fileinfo[nr].bytesinlastblock=i;
 fileinfo[nr].year=p[23];
 fileinfo[nr].month=p[24];
 fileinfo[nr].day=p[25];
 fileinfo[nr].direntry=nr;
}

/****************************************************************************/
/* Set fileinfo date to system date                                         */
/****************************************************************************/
static void SetDate (byte *p)
{
 p[23]=year-1900;
 p[24]=month+1;
 p[25]=day+1;
}

/****************************************************************************/
/* Read and interpret directory                                             */
/****************************************************************************/
static int ReadDirectory (void)
{
 int i,j;
 byte buf[1024];
 i=ReadBlock(1,buf,infile);
 if (i<0) return i;
 dirsize=buf[12]&127;
 if (buf[13]!=0x55 || buf[14]!=0xAA ||
     buf[15]!=0x00 || buf[16]!=0xFF)
  return -100;  /* Not an EOS volume */
 volumesize=buf[17]+buf[18]*256+buf[19]*256*256+buf[20]*256*256*256;
 volumesize&=imagesizemask;
 memcpy (volume_info,buf,26);
 for (i=1;i<MAX_FILE;++i) GetFileInfo (i,buf+i*26);
 for (j=1;j<dirsize;++j)
 {
  i=ReadBlock(j+1,buf,infile);
  if (i<0) return i;
  for (i=0;i<MAX_FILE;++i) GetFileInfo (i+j*MAX_FILE,buf+i*26);
 }
 return 0;
}

/****************************************************************************/
/* Get attribute priority, used for directory sorting                       */
/****************************************************************************/
static int GetAttrPriority (fileinfo_t *f)
{
 int a,r;
 a=f->attr;
 r=0;
 if (a&1) r|=128;                       /* no file                          */
 if (a&4) r|=64;                        /* deleted                          */
 if (a&8) r|=1;                         /* user                             */
 return a;
}

/****************************************************************************/
/* Compare two fileinfo structures                                          */
/****************************************************************************/
static int cmp_fileinfo (const void *a,const void *b)
{
 fileinfo_t *f_a;
 fileinfo_t *f_b;
 int i;
 f_a=(fileinfo_t *)a;
 f_b=(fileinfo_t *)b;
 if (!f_b->filename[0])
  if (!f_a->filename[0])
   return 0;
  else
   return -1;
 if (!f_a->filename[0])
  return 1;
 i=GetAttrPriority(f_a)-GetAttrPriority(f_b);
 if (i) return i;
 return strcmp (f_a->filename,f_b->filename);
}

/****************************************************************************/
/* Sort the directory for printing                                          */
/****************************************************************************/
static void SortDirectory (void)
{
 qsort (fileinfo+3,MAX_DIR-3,sizeof(fileinfo_t),cmp_fileinfo);
}

/****************************************************************************/
/* Undelete given file nr                                                   */
/****************************************************************************/
static int UndeleteFileNr (int nr)
{
 int blocknr,filenr,i;
 printf ("Undeleting %s... ",fileinfo[nr].filename);
 fflush (stdout);
 blocknr=(fileinfo[nr].direntry/MAX_FILE)+1;
 filenr=fileinfo[nr].direntry%MAX_FILE;
 i=ReadBlock(blocknr,buffer,infile);
 if (i<0)
 {
  puts ("FAILED: Cannot read file info structure");
  return i;
 }
 buffer[filenr*26+12]&=(~0x04);
 i=WriteBlock(blocknr,buffer,outfile);
 if (i<0)
 {
  puts ("FAILED: Cannot write file info structure");
  return i;
 }
 puts ("OK");
 return 0;
}

/****************************************************************************/
/* Delete given file nr                                                     */
/****************************************************************************/
static int DeleteFileNr (int nr)
{
 int blocknr,filenr,i;
 printf ("Deleting %s... ",fileinfo[nr].filename);
 fflush (stdout);
 blocknr=(fileinfo[nr].direntry/MAX_FILE)+1;
 filenr=fileinfo[nr].direntry%MAX_FILE;
 i=ReadBlock(blocknr,buffer,infile);
 if (i<0)
 {
  puts ("FAILED: Cannot read file info structure");
  return i;
 }
 buffer[filenr*26+12]|=0x04;
 i=WriteBlock(blocknr,buffer,outfile);
 if (i<0)
 {
  puts ("FAILED: Cannot write file info structure");
  return i;
 }
 puts ("OK");
 return 0;
}

/****************************************************************************/
/* Check if given file nr is to be printed                                  */
/****************************************************************************/
static int IsFile (int nr)
{
 /* Check if file is deleted (attribute flag 0x04) or system (flag 0x08) */
 if (!include_deleted)
  if (fileinfo[nr].attr&0x04) return 0;
 if (!include_system)
  if (fileinfo[nr].attr&0x08) return 0;
 /* attribute flag 0x01: No file (BLOCKS_LEFT) */
 if (fileinfo[nr].attr&1) return 0;
 /* Check for valid file length */
 if (!((keep_reserved)? fileinfo[nr].reservedblocks:fileinfo[nr].usedblocks))
  return 0;
 return 1;
}

/****************************************************************************/
/* Print information of given file nr                                       */
/****************************************************************************/
static int PrintFile (int nr)
{
 int len;
 if (!IsFile(nr)) return 0;
 len=fileinfo[nr].usedblocks*1024+fileinfo[nr].bytesinlastblock-1024;
 printf ("%14s  type %c ($%02X)  %8s  %8s  %8u byte%s",
         fileinfo[nr].filename,
         isprint(fileinfo[nr].type)? fileinfo[nr].type:'.',
         fileinfo[nr].type&255,
         (fileinfo[nr].attr&4)? "deleted":"",
         (fileinfo[nr].attr&8)? "system":"user",
         len,
         (len==1)? "":"s");
 /* Only print date if it's a valid one */
 if (fileinfo[nr].day>0 && fileinfo[nr].day<32 &&
     fileinfo[nr].month>0 && fileinfo[nr].month<13)
  printf ("  %02u-%02u-%04u\n",
          fileinfo[nr].day,fileinfo[nr].month,fileinfo[nr].year+1900);
 else
  printf ("  date unknown\n");
 return 1;
}

/****************************************************************************/
/* Print a directory listing                                                */
/****************************************************************************/
static void PrintDirectory (char *s)
{
 int i,count;
 printf ("Directory of ADAM image %s:\n",s);
 for (i=0,count=0;i<MAX_DIR;++i) count+=PrintFile(i);
 if (!count) puts ("No file found");
 for (i=0;i<MAX_DIR;++i)
  if (fileinfo[i].attr&1)
  {
   printf ("%uK free\n",fileinfo[i].reservedblocks);
   break;
  }
}

/****************************************************************************/
/* Transfer given file to home computer                                     */
/****************************************************************************/
static int TransferFileNr (int nr,char *filename)
{
 FILE *f;
 int a,i;
 printf ("Writing %s ",fileinfo[nr].filename);
 if (!filename) filename=fileinfo[nr].filename;
 else printf ("to %s ",filename);
 printf ("...");
 fflush (stdout);
 f=fopen (filename,"wb");
 if (!f) { puts ("CANNOT OPEN OUTPUT FILE"); return -1; };
 for (i=0;i<fileinfo[nr].usedblocks;++i)
 {
  a=ReadBlock(fileinfo[nr].startblock+i,buffer,infile);
  if (a<0) { puts(a==-1? "SEEK ERROR":"READ ERROR"); fclose(f); return a; }
  a=fwrite(buffer,(i==fileinfo[nr].usedblocks-1 &&
                   !ignore_bytes_in_last_block)?
           fileinfo[nr].bytesinlastblock:1024,1,f);
  if (!a) { puts ("WRITE ERROR"); fclose(f); return -3; }
 }
 fclose (f);
 puts ("OK");
 return 0;
}

/****************************************************************************/
/* Copy given file from one image to another                                */
/****************************************************************************/
static int CopyFileNr (int nr)
{
 int a,i,block_count,block_nr,dir_block_nr,dir_file_nr;
 byte *dir_ptr;
 printf ("Writing %s... ",fileinfo[nr].filename);
 fflush (stdout);
 block_count=(keep_reserved)?
             fileinfo[nr].reservedblocks:fileinfo[nr].usedblocks;
 block_nr=(fileinfo[nr].startblock<=1)?
          fileinfo[nr].startblock:first_free_block;
 if (block_nr!=1)
 {
  for (i=0;i<block_count;++i)
  {
   a=ReadBlock(fileinfo[nr].startblock+i,buffer,infile);
   if (a<0) { puts(a==-1? "SEEK ERROR":"READ ERROR"); return a; }
   a=WriteBlock(block_nr+i,buffer,outfile);
   if (a<0) { puts(a==-1? "SEEK ERROR":"WRITE ERROR"); return a; }
  }
  if (block_nr!=0) first_free_block+=block_count;
 }
 dir_block_nr=(first_free_dir_entry/MAX_FILE)+1;
 dir_file_nr=first_free_dir_entry%MAX_FILE;
 i=ReadBlock(dir_block_nr,buffer,outfile);
 if (i<0)
 {
  puts ("FAILED: Cannot read file info structure");
  return i;
 }
 dir_ptr=buffer+dir_file_nr*26;
 memcpy (dir_ptr,fileinfo[nr].realname,12);
 dir_ptr[12]=fileinfo[nr].attr;
 dir_ptr[13]=block_nr&255;
 dir_ptr[14]=(block_nr>>8)&255;
 dir_ptr[15]=(block_nr>>16)&255;
 dir_ptr[16]=(block_nr>>24)&255;
 dir_ptr[17]=block_count&255;
 dir_ptr[18]=(block_count>>8)&255;
 dir_ptr[19]=fileinfo[nr].usedblocks&255;
 dir_ptr[20]=(fileinfo[nr].usedblocks>>8)&255;
 dir_ptr[21]=fileinfo[nr].bytesinlastblock&255;
 dir_ptr[22]=(fileinfo[nr].bytesinlastblock>>8)&255;
 dir_ptr[23]=fileinfo[nr].year;
 dir_ptr[24]=fileinfo[nr].month;
 dir_ptr[25]=fileinfo[nr].day;
 i=WriteBlock(dir_block_nr,buffer,outfile);
 if (i<0)
 {
  puts ("FAILED: Cannot write file info structure");
  return i;
 }
 puts ("OK");
 first_free_dir_entry++;
 return 0;
}

/****************************************************************************/
/* Get nr. coresponding to given filename                                   */
/****************************************************************************/
static int GetFileNr (char *filename)
{
 int i;
 for (i=0;i<MAX_DIR;++i)
  if (!strcmp(filename,fileinfo[i].filename)) return i;
 return -1;
}

/****************************************************************************/
/* Transfer file to home computer                                           */
/****************************************************************************/
static int TransferFile (char *filename)
{
 int i;
 char *p,s[256];
 p=strchr (filename,',');
 if (p)
 {
  strcpy (s,p+1);
  *p='\0';
 }
 else strcpy (s,filename);
 i=GetFileNr (filename);
 if (i>=0) return TransferFileNr (i,s);
 printf ("Cannot find %s\n",filename);
 return -1;
}

/****************************************************************************/
/* Transfer given file from home computer to disk image                     */
/****************************************************************************/
static int WriteFile (char *filename,int defattr,int deftype)
{
 int attr,type,len,i,j,num_blocks;
 int dir_entry,start_block,dir_block_nr,dir_file_nr;
 char *p,*dest;
 byte *dir_ptr,dest_buf[10];
 FILE *f;
 byte old_fileinfo[26];
 memset (dest_buf,0,sizeof(dest_buf));
 for (i=0;i<10 && filename[i];++i) dest_buf[i]=filename[i];
 dest=dest_buf;
 attr=defattr;
 type=deftype;
 p=strchr(filename,',');
 if (p)
 {
  *p++='\0';
  dest=p;
  p=strchr (p,',');
  if (p)
  {
   *p++='\0';
   attr=strtoul (p,NULL,16);
   p=strchr (p,',');
   if (p)
    type=strtoul (p+1,NULL,16);
  }
 }
 if (strlen(dest)>10) dest[10]='\0';
 printf ("Writing %s to %s, attribute=$%02X, type=$%02X... ",
         filename,dest,attr,type);
 fflush (stdout);
 i=GetFileNr (dest);
 if (i>=0)
 {
  puts ("FAILED: File already exists - "
        "Please delete it and defragment the disk");
  return -3;
 }
 f=fopen (filename,"rb");
 if (!f)
 {
  puts ("FAILED: Cannot open source");
  return -4;
 }
 fseek (f,0,SEEK_END);
 len=ftell (f);
 rewind (f);
 num_blocks=len/1024; if (len%1024) ++num_blocks;
 if (num_blocks>65535 || num_blocks<=0)
 {
  puts ("FAILED: Filelength must be between 0 bytes and 65536K");
  fclose (f);
  return -5;
 }
 for (i=0;i<MAX_DIR;++i) if (fileinfo[i].attr&1) break;
 dir_entry=fileinfo[i].direntry;
 if (dir_entry+1>=MAX_FILE*dirsize)
 {
  puts ("FAILED: No free directory entry");
  fclose (f);
  return -6;
 }
 if (fileinfo[i].reservedblocks<num_blocks)
 {
  puts ("FAILED: Out of disk space");
  fclose (f);
  return -7;
 }
 start_block=fileinfo[i].startblock&imagesizemask;
 for (i=0;i<num_blocks;++i)
 {
  memset (buffer,0,1024);
  fread (buffer,1,1024,f);
  j=WriteBlock(start_block+i,buffer,outfile);
  if (j<0)
  {
   puts ("FAILED: Write error");
   fclose (f);
   return j;
  }
 }
 fclose (f);
 dir_block_nr=(dir_entry/MAX_FILE)+1;
 dir_file_nr=dir_entry%MAX_FILE;
 i=ReadBlock(dir_block_nr,buffer,infile);
 if (i<0)
 {
  puts ("FAILED: Cannot read file info structure");
  return i;
 }
 dir_ptr=buffer+dir_file_nr*26;
 memcpy (old_fileinfo,dir_ptr,26);
 memset (dir_ptr,' ',12);
 for (i=0;i<10 && dest[i];++i) dir_ptr[i]=dest[i];
 dir_ptr[i]=type; dir_ptr[i+1]=3;
 dir_ptr[12]=attr;
 dir_ptr[13]=start_block&255;
 dir_ptr[14]=(start_block>>8)&255;
 dir_ptr[15]=(start_block>>16)&255;
 dir_ptr[16]=(start_block>>24)&255;
 dir_ptr[17]=num_blocks&255;
 dir_ptr[18]=(num_blocks>>8)&255;
 dir_ptr[19]=num_blocks&255;
 dir_ptr[20]=(num_blocks>>8)&255;
 len%=1024; if (!len) len=1024;
 dir_ptr[21]=len&255;
 dir_ptr[22]=(len>>8)&255;
 dir_ptr[23]=0;
 dir_ptr[24]=0;
 dir_ptr[25]=0;
 SetDate (dir_ptr);
 i=WriteBlock(dir_block_nr,buffer,outfile);
 if (i<0)
 {
  puts ("FAILED: Cannot write file info structure");
  return i;
 }
 dir_entry++;
 dir_block_nr=(dir_entry/MAX_FILE)+1;
 dir_file_nr=dir_entry%MAX_FILE;
 i=ReadBlock(dir_block_nr,buffer,infile);
 if (i<0)
 {
  puts ("FAILED: Cannot read file info structure");
  return i;
 }
 dir_ptr=buffer+dir_file_nr*26;
 memcpy (dir_ptr,old_fileinfo,26);
 start_block+=num_blocks;
 dir_ptr[13]=start_block&255;
 dir_ptr[14]=(start_block>>8)&255;
 dir_ptr[15]=(start_block>>16)&255;
 dir_ptr[16]=(start_block>>24)&255;
 num_blocks=dir_ptr[17]+dir_ptr[18]*256-num_blocks;
 dir_ptr[17]=num_blocks&255;
 dir_ptr[18]=(num_blocks>>8)&255;
 dir_ptr[19]=0;
 dir_ptr[20]=0;
 i=WriteBlock(dir_block_nr,buffer,outfile);
 if (i<0)
 {
  puts ("FAILED: Cannot write BLOCKS_LEFT info structure");
  return i;
 }
 puts ("OK");
 return 0;
}

/****************************************************************************/
/* Transfer all files to home computer                                      */
/****************************************************************************/
static int TransferAllFiles (void)
{
 int i;
 for (i=0;i<MAX_DIR;++i)
  if (IsFile(i)) TransferFileNr (i,NULL);
 return 0;
}

/****************************************************************************/
/* Undelete file                                                            */
/****************************************************************************/
static int UndeleteFile (char *filename)
{
 int i;
 i=GetFileNr (filename);
 if (i>=0) return UndeleteFileNr (i);
 printf ("Cannot find %s\n",filename);
 return -1;
}

/****************************************************************************/
/* Delete file                                                              */
/****************************************************************************/
static int DeleteFile (char *filename)
{
 int i;
 i=GetFileNr (filename);
 if (i>=0) return DeleteFileNr (i);
 printf ("Cannot find %s\n",filename);
 return -1;
}

/****************************************************************************/
/* Main entry for reading from disk/tape images                             */
/****************************************************************************/
static void readdisk (int argc,char *argv[])
{
 char *Options[]= { "t","d","all","f","sys","del","boot",NULL };
 char *in=NULL,*boot=NULL;
 int i,j,n;
 int misparm=0;
 int tape=0;
 int all=0;
 for (n=1,i=0;n<argc;n++)
 {
  if(*argv[n]!='-')
   switch(i++)
   {
    case 0:  in=argv[n];
             break;
    default: break;
   }
  else
  {    
   for(j=0;Options[j];j++)
    if(!strcmp(argv[n]+1,Options[j])) break;
   switch(j)
   {
    case 0:  tape=1;
             break;
    case 1:  tape=0;
             break;
    case 2:  all=1;
             break;
    case 3:  ignore_bytes_in_last_block=1;
             break;
    case 4:  include_system=1;
             break;
    case 5:  include_deleted=1;
             break;
    case 6:  n++;
             if (n<argc) boot=argv[n];
             else misparm=1;
             break;
    default: misparm=1;
             break;
   }
  }
 }
 if (misparm || !in)
 {
  printf ("Usage: readdisk [options] <source> <file[,destination]>\n"
          "                [file[,destination]...]\n"
          "Available options are:\n"
          "-t           - Read from a tape image\n"
          "-d           - Read from a disk image [default]\n"
          "-sys         - Include system files\n"
          "-del         - Include deleted files\n"
          "-all         - Transfer all files\n"
          "-f           - Ignore bytes in last block field\n"
          "-boot <file> - Write boot block to specified file\n");
  fatal (1);
 }
 interleave=(tape)? tapeinterleave:diskinterleave;
 infile=fopen (in,"rb");
 outfile=NULL;
 if (!infile)
 {
  printf ("Cannot open %s\n",in);
  fatal (1);
 }
 if ((i=ReadDirectory())<0)
 {
  printf ("Cannot load directory: ");
  switch (i)
  {
   case -100: puts ("Not an EOS volume");
              break;
   case -1:   puts ("Seek error");
              break;
   default:   puts ("Read error");
  }
  fatal (2);
 }
 SortDirectory ();
 PrintDirectory (in);
 if (boot)
 {
  printf ("Writing boot block to %s... ",boot);
  fflush (stdout);
  i=ReadBlock (0,buffer,infile);
  if (i<0)
   puts ("FAILED - Cannot load boot block");
  else
  {
   FILE *f;
   f=fopen (boot,"wb");
   if (!f)
    puts ("FAILED - Cannot open file");
   else
   {
    puts ((fwrite(buffer,1,1024,f)==1024)? "OK":"FAILED - Write error");
    fclose (f);
   }
  }
 }
 if (!all)
 {
  for (n=1,i=0;n<argc;n++)
  {
   if(*argv[n]!='-')
    switch(i++)
    {
     case 0:  break;
     default: TransferFile (argv[n]);
              break;
    }
   else
   {    
    for(j=0;Options[j];j++)
     if(!strcmp(argv[n]+1,Options[j])) break;
    switch(j)
    {
     case 6:
              n++;
     default:
              break;
    }
   }
  }
 }
 else
  TransferAllFiles ();
 fatal (0);
}

/****************************************************************************/
/* Main entry for deleting and undeleting from disk/tape images             */
/****************************************************************************/
static void deldisk_common (int argc,char *argv[],int mode)
{
 char *Options[]= { "t","d",NULL };
 char *in=NULL;
 int i,j,n;
 int misparm=0;
 int tape=0;
 for (n=1,i=0;n<argc;n++)
 {
  if(*argv[n]!='-')
   switch(i++)
   {
    case 0:  in=argv[n];
             break;
    default: break;
   }
  else
  {    
   for(j=0;Options[j];j++)
    if(!strcmp(argv[n]+1,Options[j])) break;
   switch(j)
   {
    case 0:  tape=1;
             break;
    case 1:  tape=0;
             break;
    default: misparm=1;
             break;
   }
  }
 }
 if (misparm || !in)
 {
  if (mode)
  {
   printf ("Usage: deldisk [options] <source> <file> [file...]\n"
           "Available options are:\n"
           "-t          - Delete files from a tape image\n"
           "-d          - Delete files from a disk image [default]\n");
  }
  else
  {
   printf ("Usage: undeldsk [options] <source> <file> [file...]\n"
           "Available options are:\n"
           "-t          - Undelete files from a tape image\n"
           "-d          - Undelete files from a disk image [default]\n");
  }
  fatal (1);
 }
 interleave=(tape)? tapeinterleave:diskinterleave;
 infile=fopen (in,"r+b");
 outfile=infile;
 if (!infile)
 {
  printf ("Cannot open %s\n",in);
  fatal (1);
 }
 if ((i=ReadDirectory())<0)
 {
  printf ("Cannot load directory: ");
  switch (i)
  {
   case -100: puts ("Not an EOS volume");
              break;
   case -1:   puts ("Seek error");
              break;
   default:   puts ("Read error");
  }
  fatal (2);
 }
 SortDirectory ();
 PrintDirectory (in);
 for (n=1,i=0;n<argc;n++)
 {
  if(*argv[n]!='-')
   switch(i++)
   {
    case 0:  break;
    default: (mode)? DeleteFile(argv[n]) : UndeleteFile(argv[n]);
             break;
   }
 }
 fatal (0);
}

/****************************************************************************/
/* Main entry for undeleting from disk/tape images                          */
/****************************************************************************/
static void undeldisk (int argc,char *argv[])
{
 deldisk_common (argc,argv,0);
}

/****************************************************************************/
/* Main entry for deleting from disk/tape images                            */
/****************************************************************************/
static void deldisk (int argc,char *argv[])
{
 deldisk_common (argc,argv,1);
}

/****************************************************************************/
/* Main entry for writing to disk/tape images                               */
/****************************************************************************/
static void writedisk (int argc,char *argv[])
{
 char *Options[]= { "t","d","boot","attr","type","dir","size",NULL };
 char *in=NULL,*boot=NULL;
 int i,j,n,attr=0x10,type='A',dirblocks=0,volsize=0;
 int misparm=0;
 int tape=0;
 FILE *f;
 for (n=1,i=0;n<argc;n++)
 {
  if(*argv[n]!='-')
   switch(i++)
   {
    case 0:  in=argv[n];
             break;
    default: break;
   }
  else
  {    
   for(j=0;Options[j];j++)
    if(!strcmp(argv[n]+1,Options[j])) break;
   switch(j)
   {
    case 0:  tape=1;
             break;
    case 1:  tape=0;
             break;
    case 2:  n++;
             if (n<argc) boot=argv[n];
             else misparm=1;
             break;
    case 3:  n++;
             if (n<argc) attr=strtoul(argv[n],NULL,16);
             else misparm=1;
             break;
    case 4:  n++;
             if (n<argc) type=strtoul(argv[n],NULL,16);
             else misparm=1;
             break;
    case 5:  n++;
             if (n<argc) dirblocks=atoi(argv[n]);
             else misparm=1;
             break;
    case 6:  n++;
             if (n<argc) volsize=atoi(argv[n]);
             else misparm=1;
             break;
    default: misparm=1;
             break;
   }
  }
 }
 if (misparm || !in)
 {
  printf
  (
   "Usage: wrdisk [options] <source> <file[,destination[,attribute[,type]]]...>\n"
   "Available options are:\n"
   "-t            - Write to a tape image\n"
   "-d            - Write to a disk image [default]\n"
   "-boot <file>  - Write new boot block\n"
   "-attr <value> - Specify default attribute in hex [10]\n"
   "-type <value> - Specify default file type in hex [41]\n"
   "-dir <value>  - Specify number of directory blocks if writing a new image\n"
   "-size <value> - Specify size of image in kilobytes if writing a new image\n"
  );
  fatal (1);
 }
 if (volsize<32 || volsize>65535) volsize=(tape)? 256:160;
 if (!dirblocks || (dirblocks>volsize-2) || dirblocks>127) dirblocks=1;
 interleave=(tape)? tapeinterleave:diskinterleave;
 include_system=1;
 include_deleted=1;
 infile=fopen (in,"a+b");
 outfile=infile;
 if (!infile)
 {
  printf ("Cannot open %s\n",in);
  fatal (1);
 }
 fseek (infile,0,SEEK_END);
 if (ftell(infile)==0)
 {
  printf ("Writing new EOS %s image %s, size=%uK, %u directory block(s)... ",
          (tape)? "tape":"disk",in,volsize,dirblocks);
  fflush (stdout);
  newdisk[0x40c]=dirsize|0x80;                  /* Nr. of directory blocks  */
  newdisk[0x411]=volsize&255;                   /* Volume size              */
  newdisk[0x412]=(volsize>>8)&255;
  newdisk[0x45f]=(volsize-2)&255;               /* BLOCKS_LEFT size         */
  newdisk[0x460]=((volsize-2)>>8)&255;
  SetDate (newdisk+0x400+26);                   /* BOOT date                */
  SetDate (newdisk+0x400+52);                   /* DIRECTORY date           */
  if (WriteBlock(0,newdisk+0,outfile)<0)
  {
   puts ("FAILED");
   fatal (2);
  }
  if (WriteBlock(1,newdisk+1024,outfile)<0)
  {
   puts ("FAILED");
   fatal (2);
  }
  memset (buffer,0,1024);
  for (i=2;i<volsize;++i)
  {
   if (WriteBlock(i,buffer,outfile)<0)
   {
    puts ("FAILED");
    fatal (2);
   }
  }
  puts ("OK");
 }
 if ((i=ReadDirectory())<0)
 {
  printf ("Cannot load directory: ");
  switch (i)
  {
   case -100: puts ("Not an EOS volume");
              break;
   case -1:   puts ("Seek error");
              break;
   default:   puts ("Read error");
  }
  fatal (3);
 }
 SortDirectory ();
 PrintDirectory (in);
 if (boot)
 {
  printf ("Loading boot image %s... ",boot);
  fflush (stdout);
  f=fopen (boot,"rb");
  if (!f)
   puts ("FAILED");
  else
  {
   memset (buffer,0,1024);
   i=fread (buffer,1,1024,f);
   printf ("%u bytes loaded. Writing new boot image... ",i);
   fflush (stdout);
   if (WriteBlock(0,buffer,outfile)<0) puts ("FAILED");
   else puts ("OK");
   fclose (f);
  }
 }
 for (n=1,i=0;n<argc;n++)
 {
  if(*argv[n]!='-')
   switch(i++)
   {
    case 0:  break;
    default: if (i>1) ReadDirectory ();
             WriteFile (argv[n],attr,type);
             break;
   }
  else
  {    
   for(j=0;Options[j];j++)
    if(!strcmp(argv[n]+1,Options[j])) break;
   switch(j)
   {
    case 2: case 3: case 4: case 5: case 6:
     n++;
    default:
     break;
   }
  }
 }
 fatal (0);
}

/****************************************************************************/
/* Main entry for defragging disk/tape images                               */
/****************************************************************************/
static void defragdisk (int argc,char *argv[])
{
 char *Options[]= { "t","d","del","res",NULL };
 char *in=NULL;
 int i,j,n;
 int misparm=0;
 int tape=0;
 int dir_block_nr,dir_file_nr;
 byte *dir_ptr;
 for (n=1,i=0;n<argc;n++)
 {
  if(*argv[n]!='-')
   switch(i++)
   {
    case 0:  in=argv[n];
             break;
    default: misparm=1;
             break;
   }
  else
  {    
   for(j=0;Options[j];j++)
    if(!strcmp(argv[n]+1,Options[j])) break;
   switch(j)
   {
    case 0:  tape=1;
             break;
    case 1:  tape=0;
             break;
    case 2:  include_deleted=1;
             break;
    case 3:  keep_reserved=1;
             break;
    default: misparm=1;
             break;
   }
  }
 }
 if (misparm || !in)
 {
  printf ("Usage: defrdisk [options] <source>\n"
          "Available options are:\n"
          "-t          - Defragment tape image\n"
          "-d          - Defragment disk image [default]\n"
          "-del        - Keep deleted files\n"
          "-res        - Keep reserved blocks\n");
  fatal (1);
 }
 interleave=(tape)? tapeinterleave:diskinterleave;
 include_system=1;
 infile=fopen (in,"a+b");
 if (!infile)
 {
  printf ("Cannot open %s\n",in);
  fatal (1);
 }
 if ((i=ReadDirectory())<0)
 {
  printf ("Cannot load directory: ");
  switch (i)
  {
   case -100: puts ("Not an EOS volume");
              break;
   case -1:   puts ("Seek error");
              break;
   default:   puts ("Read error");
  }
  fatal (2);
 }
 SortDirectory ();
 PrintDirectory (in);
 tmpnam (tempfile);
 outfile=fopen (tempfile,"a+b");
 if (!outfile)
 {
  printf ("Cannot open temporary file %s\n",tempfile);
  fatal (3);
 }
 memcpy (buffer,volume_info,26);
 memset (buffer+26,0,1024-26);
 for (i=0;i<dirsize;++i)
 {
  j=WriteBlock (i+1,buffer,outfile);
  if (j<0)
  {
   printf ("Cannot write directory block %u\n",i+1);
   fatal (4);
  }
  memset (buffer,0,1024);
 }
 first_free_dir_entry=1;
 first_free_block=1+dirsize;
 for (i=0;i<MAX_DIR;++i)
  if (IsFile(i))
  {
   if (CopyFileNr(i)<0) fatal (5);
  }
 printf ("Writing BLOCKS_LEFT... ");
 fflush (stdout);
 dir_block_nr=(first_free_dir_entry/MAX_FILE)+1;
 dir_file_nr=first_free_dir_entry%MAX_FILE;
 i=ReadBlock(dir_block_nr,buffer,outfile);
 if (i<0)
 {
  puts ("FAILED: Cannot read file info structure");
  fatal (6);
 }
 dir_ptr=buffer+dir_file_nr*26;
 memcpy (dir_ptr,"BLOCKS LEFT\03",12);
 dir_ptr[12]=0x01;
 dir_ptr[13]=first_free_block&255;
 dir_ptr[14]=(first_free_block>>8)&255;
 dir_ptr[15]=(first_free_block>>16)&255;
 dir_ptr[16]=(first_free_block>>24)&255;
 dir_ptr[17]=(volumesize-first_free_block)&255;
 dir_ptr[18]=((volumesize-first_free_block)>>8)&255;
 dir_ptr[19]=0;
 dir_ptr[20]=0;
 dir_ptr[21]=0x00;
 dir_ptr[22]=0x04;
 dir_ptr[23]=0;
 dir_ptr[24]=0;
 dir_ptr[25]=0;
 i=WriteBlock(dir_block_nr,buffer,outfile);
 if (i<0)
 {
  puts ("FAILED: Cannot write file info structure");
  fatal (6);
 }
 memset (buffer,0,1024);
 printf ("OK\nClearing unused blocks... ");
 fflush (stdout);
 for (i=first_free_block;i<volumesize;++i)
  if (WriteBlock(i,buffer,outfile)<0)
  {
   puts ("FAILED");
   fatal (7);
  }
 printf ("OK\nCopying temporary file %s to %s... ",tempfile,in);
 fflush (stdout);
 rewind (infile);
 rewind (outfile);
 while ((i=fread(buffer,1,1024,outfile))!=0)
  fwrite(buffer,1,i,infile);
 puts ("OK");
 fatal (0);
}

struct program_struct
{
 char *name;
 void (*fn) (int argc,char *argv[]);
};

/****************************************************************************/
/* Get system date and initialise day, month and year                       */
/****************************************************************************/
static void GetSystemDate (void)
{
 time_t timer;
 struct tm *tblock;
 timer=time(NULL);
 tblock=localtime(&timer);
 day=tblock->tm_mday-1;
 month=tblock->tm_mon;
 year=tblock->tm_year+1900;
}

/****************************************************************************/
/* Program entry. Get program name (readdisk, deldisk, etc.) and call       */
/* appropriate function                                                     */
/****************************************************************************/
int main (int argc,char *argv[])
{
 struct program_struct programs[]=
 {
  { "deldisk"   ,deldisk },
  { "undeldsk"  ,undeldisk },
  { "readdisk"  ,readdisk },
  { "wrdisk"    ,writedisk },
  { "defrdisk"  ,defragdisk },
  { NULL,NULL },
 };
 struct program_struct options[]=
 {
  { "--delete"   ,deldisk },
  { "--undelete" ,undeldisk },
  { "--read"     ,readdisk },
  { "--write"    ,writedisk },
  { "--defrag"   ,defragdisk },
  { NULL,NULL },
 };
 char *p,*q,*progname;
 int i;
 printf ("ADAM EOS disk image utilities\n"
         "Copyright (C) Marcel de Kogel 1996,1997\n");
#ifdef MSDOS
 strlwr (argv[0]);
#endif
 GetSystemDate ();
 p=argv[0];
 q=strchr(p,'/');
 while (q)                              /* get last '/'                     */
 {
  p=++q;
  q=strchr(q,'/');
 };
 progname=p;
 q=NULL;
 while ((p=strchr(p,'.'))!=NULL)        /* get last '.'                     */
 {
  q=p;
  ++p;
 }
 if (q) *q='\0';                        /* remove extension                 */
 for (i=0;programs[i].name;++i)
  if (!strcmp(progname,programs[i].name)) (*programs[i].fn)(argc,argv);
 /* we weren't started via a symlink, so check first command line option */
 if (argc>1)
 {
  for (i=0;options[i].name;++i)
   if (!strcmp(argv[1],options[i].name)) (*options[i].fn)(argc-1,argv+1);
 }
 printf ("Usage: eosdisk [mode] [options]\n"
         "Available modes are:\n"
         "--read      - Read from EOS image\n"
         "--write     - Write to EOS image\n"
         "--delete    - Delete files from EOS image\n"
         "--undelete  - Undelete files from EOS image\n"
         "--defrag    - Defragment EOS image\n"
         "Available options depend on mode chosen\n");
 return 0;
}
